(defvar mixed-coding-separator-alist
  '(("^GB2312.*" . euc-cn)
    ("^BIG5.*" . big5)
    ("^JIS-euc.*" . euc-jp)))

(defconst mixed-coding-separator-regexp
  (mapconcat
   #'(lambda (x) (concat "\\(" (car x) "\\)"))
   mixed-coding-separator-alist
   "\\|"))

(defconst mixed-coding-safe-charsets
  (let ((l nil))
    (dolist (elt mixed-coding-separator-alist)
      (let ((charsets (coding-system-get (cdr elt) 'safe-charsets)))
	(if (not l)
	    (setq l charsets)
	  (dolist (e charsets)
	    (or (memq e l)
		(setq l (cons e l)))))))
    l))

(defun mixed-coding-post-read-converion (len)
  (save-restriction
    (narrow-to-region (point) (+ (point) len))
    (let ((pos (point))
	  (coding nil))
      (while (re-search-forward mixed-coding-separator-regexp nil 'move)
	(let ((separator (match-string 0)))
	  (put-text-property (match-beginning 0) (match-end 0)
			     'face 'highlight)
	  (put-text-property (match-beginning 0) (match-end 0)
			     'intangible t)
	  (if coding
	      (decode-coding-region pos (match-beginning 0) coding))
	  (setq pos (point) coding nil)
	  (let ((tail mixed-coding-separator-alist))
	    (while (and (not coding) tail)
	      (if (string-match (car (car tail)) separator)
		  (setq coding (cdr (car tail)))
		(setq tail (cdr tail)))))))
      (if coding
	  (decode-coding-region pos (point) coding)))
    (goto-char (point-min))
    (setq last-coding-system-used 'mixed-coding)
    (- (point-max) (point))))

(defun mixed-coding-pre-write-converion (from to)
  (let ((buf (current-buffer)))
    (set-buffer (generate-new-buffer " *temp*"))
    (if (stringp from)
	(insert from)
      (insert-buffer-substring buf from to))
    (goto-char (point-min))
    (let ((pos (point))
	  (coding nil))
      (while (re-search-forward mixed-coding-separator-regexp nil 'move)
	(let ((separator (match-string 0)))
	  (if coding
	      (encode-coding-region pos (point) coding))
	  (setq pos (point) coding nil)
	  (let ((tail mixed-coding-separator-alist))
	    (while (and (not coding) tail)
	      (if (string-match (car (car tail)) separator)
		  (setq coding (cdr (car tail)))
		(setq tail (cdr tail)))))))
      (if coding
	  (encode-coding-region pos (point) coding)))
    (setq last-coding-system-used 'mixed-coding)
    nil))

(make-coding-system
 'mixed-coding 0 ?M
 "Handle mixed codings."
 nil
 `((safe-charsets . ,mixed-coding-safe-charsets)
   (post-read-conversion . mixed-coding-post-read-converion)
   (pre-write-conversion . mixed-coding-pre-write-converion)))

